/*
 *  linux/ibcs/open.c
 *
 *  Copyright (C) 1993  Joe Portman (baron@hebron.connected.com)
 *  Copyright (C) 1993, 1994  Drew Sullivan (re-worked for iBCS2)
 *
 * $Id: open.c,v 1.1.1.2 1994/02/23 17:22:52 mike Exp $
 * $Source: /var/CVS/ibcs/open.c,v $
 */

/* Keep track of which struct definition we really want */
#include <linux/vfs.h>
#include <linux/types.h>
#include <linux/utime.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/tty.h>
#include <linux/time.h>
#include <linux/malloc.h>

#include <asm/segment.h>

#include <ibcs/ibcs.h>

#ifdef __cplusplus
extern "C" 
#endif
int ibcs_statfs(const char * path, struct ibcs_statfs * buf)
{
	struct inode * inode;
	struct statfs lxstat;
	struct ibcs_statfs ibcsstat;
	int error;
	int old_fs;

	error = verify_area(VERIFY_WRITE, buf, sizeof(struct ibcs_statfs));
	if (error)
		return error;
	error = namei(path,&inode);
	if (error)
		return error;
	if (!inode->i_sb->s_op->statfs) {
		iput(inode);
		return -ENOSYS;
	}
	old_fs = get_fs();
	set_fs(get_ds());

	inode->i_sb->s_op->statfs(inode->i_sb, &lxstat);

	set_fs(old_fs);
	iput(inode);

	ibcsstat.f_type = lxstat.f_type;
	ibcsstat.f_bsize = lxstat.f_bsize;
	ibcsstat.f_frsize = lxstat.f_bsize;
	ibcsstat.f_blocks = lxstat.f_blocks;
	ibcsstat.f_bfree = lxstat.f_bfree;
	ibcsstat.f_files = lxstat.f_files;
	ibcsstat.f_ffree = lxstat.f_ffree;
	memset(ibcsstat.f_fname, 0, sizeof(ibcsstat.f_fname));
	memset(ibcsstat.f_fpack, 0, sizeof(ibcsstat.f_fpack));

	/* Finally, copy it to the user's buffer */
	memcpy_tofs(&ibcsstat, buf, sizeof(ibcsstat));
	return 0;
}

#ifdef __cplusplus
extern "C" 
#endif
int ibcs_fstatfs(unsigned int fd, struct ibcs_statfs * buf)
{
	struct inode * inode;
	struct file * file;
	struct statfs lxstat;
	struct ibcs_statfs ibcsstat;
	int error;
	int old_fs;

	error = verify_area(VERIFY_WRITE, buf, sizeof(struct ibcs_statfs));
	if (error)
		return error;
	if (fd >= NR_OPEN || !(file = current->filp[fd]))
		return -EBADF;
	if (!(inode = file->f_inode))
		return -ENOENT;
	if (!inode->i_sb->s_op->statfs)
		return -ENOSYS;

	old_fs = get_fs();
	set_fs(get_ds());

	inode->i_sb->s_op->statfs(inode->i_sb, &lxstat);

	set_fs(old_fs);


	ibcsstat.f_type = lxstat.f_type;
	ibcsstat.f_bsize = lxstat.f_bsize;
	ibcsstat.f_frsize = lxstat.f_bsize;
	ibcsstat.f_blocks = lxstat.f_blocks;
	ibcsstat.f_bfree = lxstat.f_bfree;
	ibcsstat.f_files = lxstat.f_files;
	ibcsstat.f_ffree = lxstat.f_ffree;
	memset(ibcsstat.f_fname, 0, sizeof(ibcsstat.f_fname));
	memset(ibcsstat.f_fpack, 0, sizeof(ibcsstat.f_fpack));

	/* Finally, copy it to the user's buffer */
	memcpy_tofs(&ibcsstat, buf, sizeof(ibcsstat));
	return 0;
}

/*
** okay, translate ibcs open flags into linux open flags
** for ISC unix, there is a difference, so we
** are just kludging it here
*/
#define U_RDONLY	00000
#define U_WRONLY	00001
#define	U_RDWR		00002
#define U_NDELAY	00004
#define	U_APPEND	00010
#define	U_SYNC 		00020
#define U_NONBLOCK	00200
#define	U_CREAT		00400
#define	U_TRUNC		01000
#define U_EXCL		02000
#define	U_NOCTTY	04000

/*
** okay, ISC unix does some nasty 
** thing with TRUNC, it assumes
** O_CREAT, I guess.
** This is an error, but we
** will kludge it here for compatibility.
*/

int ibcs_translate_open_flag(int flag)
{
	int oflag = 0;

	/* if  (flag & U_RDONLY)  oflag |= O_RDONLY;  -- RDONLY is zero */
	if  (flag & U_WRONLY) 	oflag |= O_WRONLY;
	if  (flag & U_RDWR)   	oflag |= O_RDWR;
	if  (flag & U_NDELAY) 	oflag |= O_NDELAY;
	if  (flag & U_APPEND) 	oflag |= O_APPEND;
	if  (flag & U_NONBLOCK) oflag |= O_NONBLOCK;
	if  (flag & U_SYNC) 	oflag |= O_SYNC;
	if  (flag & U_CREAT) 	oflag |= O_CREAT;
	if  (flag & U_TRUNC) 	oflag |= (O_TRUNC|O_CREAT);
	if  (flag & U_EXCL) 	oflag |= O_EXCL;
	if  (flag & U_NOCTTY) 	oflag |= O_NOCTTY;

	/* printk("ibcs: open iflag %08o oflag %08o\n",flag,oflag); */
	return oflag;
}

int ibcs_open(const char * fname, int flag, int mode)
{
	return sys_open(fname, ibcs_translate_open_flag(flag), mode);
}


/* If/when the readdir function is changed to read multiple entries
 * at once this should be updated to take advantage of the fact.
 *
 * N.B. For Linux the reclen in a dirent is the number of characters
 * in the filename, for SCO (at least) reclen is the total size of
 * the particular dirent rounded up to the next multiple of 4. The SCO
 * behaviour is faithfully emulated here.
 *
 * XXXX
 * We don't truncate long filenames at all when copying. If we meet a
 * long filename and the buffer supplied by the application simply isn't
 * big enough to hold it we'll return without filling the buffer (i.e
 * return 0). The application will see this as a (premature) end of
 * directory. Is there a work around for this at all???
 */
int ibcs_getdents(int fd, char *buf, int nbytes)
{
	int error, here, posn, reclen;
	struct file *file;
	struct dirent *d;
	int old_fs;

	error = verify_area(VERIFY_WRITE, buf, nbytes);
	if (error)
		return error;

	/* Check the file handle here. This is so we can access the current
	 * position in the file structure safely without a tedious call
	 * to sys_lseek that does nothin useful.
	 */
	if (fd >= NR_OPEN || !(file=current->filp[fd]) || !(file->f_inode))
		return -EBADF;

	/* XXXXXX
	 * The size of dirent must be explicit and not rely on the name
	 * array quietly extending past the end of the structure!
	 * It may be safer just to use get_free_page for this?
	 */
	d = (struct dirent *)kmalloc(sizeof(struct dirent), GFP_KERNEL);
	if (!d)
		return -ENOMEM;

	error = posn = reclen = 0;
	while (posn + reclen < nbytes) {
		/* Save the current position and get another dirent */
		here = file->f_pos;
		old_fs = get_fs();
		set_fs (get_ds());
		error = sys_readdir(fd, d, 1);
		set_fs(old_fs);
		if (error <= 0)
			break;

		/* If it'll fit in the buffer save it otherwise back up
		 * so it is read next time around.
		 * Oh, if we're at the beginning of the buffer there's
		 * no chance that this entry will ever fit so don't
		 * copy it and don't back off - we'll just pretend it
		 * isn't here...
		 */
		reclen = (sizeof(long) + sizeof(off_t)
			+ sizeof(unsigned short) + d->d_reclen + 1
			+ 3) & (~3);
		if (posn + reclen <= nbytes) {
			d->d_reclen = reclen;
			d->d_off = file->f_pos;
			memcpy_tofs(buf+posn, d, reclen);
			posn += reclen;
		} else if (posn) {
			sys_lseek(fd, here, 0);
		} /* else posn == 0 */
	}

	/* Loose the intermediate buffer. */
	kfree(d);

	/* If we've put something in the buffer return the byte count
	 * otherwise return the error status.
	 */
	return ((posn > 0) ? posn : error);
}
